from PyQt6.QtCore import QThread
from .live_scene import LiveScene


class RenderThread(QThread):
    """A thread to render the scene."""

    def __init__(self, scene: LiveScene, preview: bool, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.preview = preview
        self.scene = scene

    def run(self):
        self.scene.render(self.preview)



from manim import ValueTracker, Mobject, ManimColor
import numpy as np


class ColorValueTracker(ValueTracker):
    """A value tracker for color. It contains an array of 3 values and an alpha value."""

    def __init__(self, value=np.array([0, 0, 0, 1]), **kwargs):
        Mobject.__init__(self, **kwargs)
        self.set_points(np.zeros((1, 4)))
        self.set_value(value)

    def get_value(self) -> ManimColor:
        return ManimColor.from_rgba(self.points[0, :4]).to_hex(), self.points[0, 3]

    def set_value(self, value: np.ndarray):
        self.points[0, :4] = np.array(value)
        return self



from PyQt6.QtWidgets import QSlider

from manim_studio.value_trackers.int_value_tracker import IntValueTracker


class Slider(QSlider):
    """
    A slider widget to edit an integer property
    that can be set to a value between a minimum and a maximum.
    There's also a step value that can be set.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_tracker = IntValueTracker(0)
        self.valueChanged.connect(self.value_tracker.set_value)



from PyQt6.QtWidgets import QVBoxLayout, QLabel, QSlider, QColorDialog, QComboBox, QLineEdit, QTextEdit, QCheckBox, \
    QPushButton, QScrollArea, QGroupBox, QDoubleSpinBox, QHBoxLayout
from PyQt6.QtGui import QColor
from PyQt6.QtCore import QThread, pyqtSignal, Qt
import socket
import numpy as np


class ClientControls(QScrollArea):
    """A widget to display the controls of the scene in Manim Studio Client."""

    add_slider_to_client = pyqtSignal(str, str, str, str, str)
    add_color_widget_to_client = pyqtSignal(str, int, int, int, int)
    add_dropdown_to_client = pyqtSignal(str, list, str)
    add_line_edit_to_client = pyqtSignal(str, str)
    add_text_editor_to_client = pyqtSignal(str, str)
    add_checkbox_to_client = pyqtSignal(str, bool)
    add_button_to_client = pyqtSignal(str)
    add_position_control_to_client = pyqtSignal(str, np.ndarray)
    set_control_value = pyqtSignal(str, str)
    close_signal = pyqtSignal()

    def __init__(self, main, s: socket.socket, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.main = main
        self.setWindowTitle("Manim Studio Client - Controls")
        self.layout_ = QVBoxLayout()
        self.controls = {}
        self.setLayout(self.layout_)
        self.label = QLabel(text="No controls defined")
        self.layout().addWidget(self.label)
        self.s = s
        self.add_slider_to_client.connect(
            self.add_slider_to_client_command)
        self.add_color_widget_to_client.connect(
            self.add_color_widget_to_client_command)
        self.add_dropdown_to_client.connect(
            self.add_dropdown_to_client_command)
        self.add_line_edit_to_client.connect(
            self.add_line_edit_to_client_command)
        self.add_text_editor_to_client.connect(
            self.add_text_editor_to_client_command)
        self.add_checkbox_to_client.connect(
            self.add_checkbox_to_client_command)
        self.add_button_to_client.connect(
            self.add_button_to_client_command)
        self.set_control_value.connect(
            self.set_control_value_command)
        self.add_position_control_to_client.connect(
            self.add_position_control_to_client_command)
        self.close_signal.connect(self.close)
        self.update_controls_thread = UpdateControlsThread(
            self.s,
            add_slider_to_client=self.add_slider_to_client,
            add_color_widget_to_client=self.add_color_widget_to_client,
            add_dropdown_to_client=self.add_dropdown_to_client,
            add_line_edit_to_client=self.add_line_edit_to_client,
            add_text_editor_to_client=self.add_text_editor_to_client,
            add_checkbox_to_client=self.add_checkbox_to_client,
            add_button_to_client=self.add_button_to_client,
            add_position_control_to_client=self.add_position_control_to_client,
            close_signal=self.close_signal,
            set_control_value=self.set_control_value)
        self.update_controls_thread.start()

    def set_control_value_command(self, name, value):
        if name in self.controls:
            control = self.controls[name]
            if isinstance(control, QSlider):
                control.setValue(int(value))
            elif isinstance(control, QColorDialog):
                r, g, b, a = value.split(",")
                control.setCurrentColor(
                    QColor(int(r), int(g), int(b), int(a)))
            elif isinstance(control, QComboBox):
                control.setCurrentText(value)
            elif isinstance(control, QLineEdit):
                control.setText(value)
            elif isinstance(control, QTextEdit):
                control.setText(value)
            elif isinstance(control, QCheckBox):
                control.setChecked(value == "True")
            elif isinstance(control, QGroupBox):
                x, y, z = value.split(",")
                control.x_.setValue(float(x))
                control.y_.setValue(float(y))
                control.z_.setValue(float(z))
            else:
                return
        else:
            return

    def add_slider_to_client_command(self, name, min_, max_, step, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        slider = QSlider()
        slider.setOrientation(Qt.Orientation.Horizontal)
        slider.setMinimum(int(min_))
        slider.setMaximum(int(max_))
        slider.setSingleStep(int(step))
        slider.setValue(int(default))
        slider.valueChanged.connect(
            lambda value: self.s.sendall(f"set_slider_value {name} {value}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(slider)
        self.controls[name] = slider

    def add_color_widget_to_client_command(self, name, r, g, b, a):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        color_dialog = QColorDialog()
        color_dialog.setCurrentColor(
            QColor(int(r), int(g), int(b), int(a)))
        color_dialog.currentColorChanged.connect(
            lambda color: self.s.sendall(f"set_color {name} {color.red()} {color.green()} {color.blue()} {color.alpha()}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(color_dialog)
        self.controls[name] = color_dialog

    def add_dropdown_to_client_command(self, name, options, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        dropdown = QComboBox()
        dropdown.addItems(options)
        dropdown.setCurrentText(default)
        dropdown.currentTextChanged.connect(
            lambda text: self.s.sendall(f"set_dropdown {name} {text}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(dropdown)
        self.controls[name] = dropdown

    def add_line_edit_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        line_edit = QLineEdit()
        line_edit.setText(default)
        line_edit.textChanged.connect(
            lambda text: self.s.sendall(f"set_line_edit {name} {text}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(line_edit)
        self.controls[name] = line_edit

    def add_text_editor_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        text_edit = QTextEdit()
        text_edit.setText(default)
        text_edit.textChanged.connect(
            lambda text: self.s.sendall(f"set_text_edit {name} {text}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(text_edit)
        self.controls[name] = text_edit

    def add_checkbox_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        checkbox = QCheckBox()
        checkbox.setChecked(default == "True")
        checkbox.stateChanged.connect(
            lambda: self.s.sendall(f"set_checkbox {name} {checkbox.isChecked()}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(checkbox)
        self.controls[name] = checkbox

    def add_button_to_client_command(self, name):
        if self.label.isVisible():
            self.label.hide()
        name_label = QLabel(text=name)
        button = QPushButton(text=name)
        button.clicked.connect(
            lambda: self.s.sendall(f"button_clicked {name}".encode("utf-8")))
        self.layout().addWidget(name_label)
        self.layout().addWidget(button)
        self.controls[name] = button

    def close(self):
        super().close()
        self.main.close()
        self.s.close()

    def add_position_control_to_client_command(self, name, default):
        if self.label.isVisible():
            self.label.hide()
        widget = QGroupBox(name)
        widget.setLayout(QHBoxLayout())
        widget.x_ = QDoubleSpinBox()
        widget.x_.setValue(default[0])
        widget.y_ = QDoubleSpinBox()
        widget.y_.setValue(default[1])
        widget.z_ = QDoubleSpinBox()
        widget.z_.setValue(default[2])
        widget.x_.valueChanged.connect(
            lambda value: self.s.sendall(f"set_position_control_value_x {name} {value}".encode("utf-8")))
        widget.y_.valueChanged.connect(
            lambda value: self.s.sendall(f"set_position_control_value_y {name} {value}".encode("utf-8")))
        widget.z_.valueChanged.connect(
            lambda value: self.s.sendall(f"set_position_control_value_z {name} {value}".encode("utf-8")))
        widget.display_dot_checkbox = QCheckBox("Display Dot")
        widget.display_dot_checkbox.setChecked(False)
        widget.display_dot_checkbox.stateChanged.connect(
            lambda: self.s.sendall(f"set_position_control_display_dot {name} {widget.display_dot_checkbox.isChecked()}".encode("utf-8")))
        widget.layout().addWidget(widget.x_)
        widget.layout().addWidget(widget.y_)
        widget.layout().addWidget(widget.z_)
        widget.layout().addWidget(widget.display_dot_checkbox)
        self.layout().addWidget(widget)
        self.controls[name] = widget


class UpdateControlsThread(QThread):
    def __init__(self,
                 s: socket.socket,
                 add_slider_to_client: pyqtSignal,
                 add_color_widget_to_client: pyqtSignal,
                 add_dropdown_to_client: pyqtSignal,
                 add_line_edit_to_client: pyqtSignal,
                 add_text_editor_to_client: pyqtSignal,
                 add_checkbox_to_client: pyqtSignal,
                 add_button_to_client: pyqtSignal,
                 add_position_control_to_client: pyqtSignal,
                 close_signal: pyqtSignal,
                 set_control_value: pyqtSignal):
        super().__init__()
        self.s = s
        self.add_slider_to_client = add_slider_to_client
        self.add_color_widget_to_client = add_color_widget_to_client
        self.add_dropdown_to_client = add_dropdown_to_client
        self.add_line_edit_to_client = add_line_edit_to_client
        self.add_text_editor_to_client = add_text_editor_to_client
        self.add_checkbox_to_client = add_checkbox_to_client
        self.add_button_to_client = add_button_to_client
        self.add_position_control_to_client = add_position_control_to_client
        self.close_signal = close_signal
        self.set_control_value = set_control_value

    def run(self):
        self.s.sendall(b"get_controls")
        while True:
            try:
                msg = self.s.recv(1024)
            except ConnectionResetError:
                break
            command = msg.decode("utf-8")
            if command == "no_controls":
                continue
            if command.startswith("add_slider"):
                _, name, default, min_, max_, step = command.split(" ")
                self.add_slider_to_client.emit(
                    name, min_, max_, step, default)
            elif command.startswith("add_color_widget"):
                _, name, r, g, b, a = command.split(" ")
                self.add_color_widget_to_client.emit(
                    name, r, g, b, a)
            elif command.startswith("add_dropdown"):
                _, name, options, default = command.split(" ")
                options = options.split(",")
                self.add_dropdown_to_client.emit(
                    name, options, default)
            elif command.startswith("add_line_editor_widget"):
                _, name, *words = command.split(" ")
                self.add_line_edit_to_client.emit(
                    name, " ".join(words))
            elif command.startswith("add_text_editor_widget"):
                _, name, *words = command.split(" ")
                self.add_text_editor_to_client.emit(
                    name, " ".join(words))
            elif command.startswith("add_checkbox"):
                _, name, default = command.split(" ")
                self.add_checkbox_to_client.emit(
                    name, default == "True")
            elif command.startswith("add_button"):
                _, name = command.split(" ")
                self.add_button_to_client.emit(name)
            elif command.startswith("add_position_control"):
                _, name, *default = command.split(" ")
                x, y, z = default
                x = float(x)
                y = float(y)
                z = float(z)
                self.add_position_control_to_client.emit(
                    name, np.array([x, y, z]))
            elif command == "exit":
                self.close_signal.emit()
                break
            else:
                continue



from PyQt6.QtCore import pyqtSignal, QObject
import numpy as np
import socket


class Communicate(QObject):
    """A class to communicate using signals and slots."""
    update_scene = pyqtSignal(str)
    update_image = pyqtSignal(np.ndarray)
    end_scene = pyqtSignal()
    alert = pyqtSignal(Exception)
    save_snippet = pyqtSignal(str)
    next_slide = pyqtSignal()
    add_slider_to_editor = pyqtSignal(str, str, str, str, str)
    add_color_widget_to_editor = pyqtSignal(str, np.ndarray)
    add_dropdown_to_editor = pyqtSignal(str, list, str)
    add_line_edit_to_editor = pyqtSignal(str, str)
    add_text_editor_to_editor = pyqtSignal(str, str)
    add_checkbox_to_editor = pyqtSignal(str, bool)
    add_button_to_editor = pyqtSignal(str, str)
    add_position_control_to_editor = pyqtSignal(str, np.ndarray)
    save_mobject = pyqtSignal()
    load_mobject = pyqtSignal()
    set_control_value = pyqtSignal(str, str)
    add_controls_to_client = pyqtSignal(socket.socket, dict)
    add_controls_to_clients = pyqtSignal(list, dict)
    press_button = pyqtSignal(str)
    set_position_control_x = pyqtSignal(str, str)
    set_position_control_y = pyqtSignal(str, str)
    set_position_control_z = pyqtSignal(str, str)
    set_position_control_display_dot = pyqtSignal(str, str)
    show_in_status_bar = pyqtSignal(str)
    pause_scene = pyqtSignal()
    resume_scene = pyqtSignal()
    print_gui = pyqtSignal(str)
    screenshot = pyqtSignal(str)



from PyQt6.QtWidgets import QApplication
from PyQt6.QtGui import QIcon
import sys

from .editor_widget import EditorWidget
from .preview_widget import PreviewWidget
from .live_scene import LiveScene
from .communicate import Communicate
from .ai_widget import AIWidget
from .get_icon_file import get_icon_file
from .inherits_dialog import InheritsDialog


def main(scene_type=LiveScene, server=False, namespace=None, preview=False) -> None:
    """
    Run the main window of Manim Studio.

    Parameters
    ----------
    scene_type
        The base class of the scene to be rendered.
    server
        Whether to run the server.
    namespace
        The namespace of the scene to be rendered.
    preview
        Whether to preview the final result.
    """
    app = QApplication(sys.argv)
    screen_size = app.primaryScreen().size()
    w, h = screen_size.width(), screen_size.height()
    icon = QIcon(get_icon_file())
    app.setWindowIcon(icon)
    inherits_dialog = InheritsDialog(scene_type, namespace)
    inherits_dialog.exec()
    scene_type = inherits_dialog.set_scene_type_base_classes()
    communicate = Communicate()
    scene = scene_type(communicate, namespace)
    editor_window = EditorWidget(communicate, scene, server)
    preview_window = PreviewWidget(communicate, scene, (w, h), preview)
    preview_window.show()
    editor_window.show()
    if AIWidget is not None:
        ai_window = AIWidget(communicate)
        ai_window.show()
    if server:
        from .server import ManimStudioServer
        server_window = ManimStudioServer(communicate, editor_window)
        server_window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main(LiveScene)



def load_snippet(scene, file_name):
    """Load a snippet from a file and execute it."""
    with open(file_name, "r") as f:
        code = f.read()
    scope = globals()
    scope["self"] = scene
    exec(code, scope)



from PyQt6.QtWidgets import QDialog, QVBoxLayout, QGroupBox, QCheckBox
import manim
from .live_scene import LiveScene


class InheritsDialog(QDialog):
    def __init__(self, scene_type=LiveScene, namespace=None):
        """A dialog to choose base classes of the scene to be rendered."""

        super().__init__()
        self.layout_ = QVBoxLayout()
        self.setLayout(self.layout_)
        self.setWindowTitle("Manim Studio - Use base class")
        self.setGeometry(0, 0, 400, 300)
        self.group_box = QGroupBox("Choose base classes")
        self.group_box.setLayout(QVBoxLayout())
        # Allow multiple inheritance
        self.namespace = namespace.__dict__ if namespace is not None else {}
        self.namespace.update(manim.__dict__)
        for name, obj in self.namespace.items():
            if isinstance(obj, type) and issubclass(obj, manim.Scene) and obj not in [manim.Scene, LiveScene]:
                checkbox = QCheckBox(name)
                checkbox.setChecked(False)
                self.group_box.layout().addWidget(checkbox)
        self.layout_.addWidget(self.group_box)
        self.scene_type = scene_type

    def set_scene_type_base_classes(self):
        for i in range(self.group_box.layout().count()):
            checkbox = self.group_box.layout().itemAt(i).widget()
            if checkbox.isChecked():
                if manim.Scene in self.scene_type.__bases__:
                    self.scene_type.__bases__ = tuple(
                        item for item in self.scene_type.__bases__ if item != manim.Scene)
                self.scene_type.__bases__ = self.scene_type.__bases__ + \
                    (self.namespace[checkbox.text()],)
        return self.scene_type



from manim import Mobject
import dill as pickle


def load_mobject(path: str) -> Mobject:
    with open(path, "rb") as f:
        return pickle.load(f)


def save_mobject(mobject: Mobject, path: str, scope) -> None:
    scope["console"] = None  # To fix a bug
    scope["logger"] = None  # To fix a bug
    scope["error_console"] = None  # To fix a bug
    self_ = scope["self"]  # To fix a bug
    scope["self"] = None  # To fix a bug
    with open(path, "wb") as f:
        pickle.dump(mobject, f)
    scope["self"] = self_  # To fix a bug



from manim_studio.code_edit import CodeEdit
from manim_studio.value_trackers.string_value_tracker import StringValueTracker


class TextEditorWidget(CodeEdit):
    """A text editor widget to edit a string property using multiple lines."""

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_tracker = StringValueTracker("")
        self.textChanged.connect(
            lambda: self.value_tracker.set_value(self.toPlainText()))
